#!/usr/bin/env python3

# sendfilemq.py
# Python port of SendFileMQv14_bas, SendFileMQv22_bas by Martyn Hill

import os
import serial as pyserial
import sys
import time

# SBASIC standard and extended commands implemented to simplify port

def b1(i):
    return bytes([i])

def b2(i):
    if i < 0:
        return b2(65536 + i)
    return i.to_bytes(2, byteorder="big")

def b4(i):
    if i < 0:
        return b4(4294967296 + i)
    return i.to_bytes(4, byteorder="big")


def i2(two_chars):
    return 256 * two_chars[0] + two_chars[1]


ttt = 0

# omit recvTimer% arg and assume a global
def T_START():
    global ttt
    ttt = time.time() * 50


# omit recvTimer% arg and assume a global
def T_COUNT():
    global ttt
    return (time.time() * 50) - ttt


# omit recvTimer% arg and assume a global
def T_STOP():
    # no op needed
    pass


# Constants and globals

netBlkLen = 255
netHdrLen = 8

fHdrLenQL = 15

mqHdrLen = 16
mQHdr = bytearray(mqHdrLen)

maxMID = 8191
minMID = 1

mQErrorMask = 2 ** 15
mQLongMask = 2 ** 14
mQReplyMask = 2 ** 13

mQMIDMask = 65536 + ~(mQErrorMask | mQLongMask | mQReplyMask)

mQSignature = b"%M$Q"

# *** Net Profiles allow the Timing Constants to be dynamically updated (as of QLUB v2.2a)
netProfile = [
    # *** 0: Suited for QL
    # NB AH ROM has different timing characteristics
    # b"\x00\xC8\x4E\x20\x00\x32\x01\xDB\x0B\xB8\x00\x06\x01\xF4\x13\x88\x00\xC8\x00\x00\x00\x8C\x00\xB4\x9C\x40\x05\x00\x01\x0E\x00\xB4",
    b"\x00\xC8\x4E\x20\x00\x32\x01\xDB\x0B\xB8\x00\x06\x01\xF4\x13\x88\x00\xC8\x00\x00\x00\x8C\x00\xB8\x9C\x40\x05\x00\x01\x0E\x00\xB4",
    # *** 1: Suited for ZX Spectrum - only ndt_send (item 12) and ndt_rdly (item 15) seem to need adjustment
    b"\x00\xC8\x4E\x20\x00\x32\x01\xDB\x0B\xB8\x00\x06\x01\xF4\x13\x88\x00\xC8\x00\x00\x00\x8C\x00\xC0\x9C\x40\x05\x00\x01\x27\x00\xB4",
]

nextMID = minMID

def getNextMID():
    global nextMID
    thisMID = nextMID
    nextMID += 1
    if nextMID > maxMID:
        nextMID = minMID
    return thisMID


def CalcChkSum(p, MODflag):
    chk = 0
    for i in p:
        chk += i
        if MODflag:
            chk %= 256
    return chk


def prepareMQHdr(mID, mLongFlag, mDev, mAct, mPar0, mParErr, mLenPar2):
    mID = b2(mID)
    if mLongFlag:
        mID = bytes([mID[0] + 64, mID[1]])
    mPar0s = b2(mPar0)
    mParErrs = b2(mParErr)
    mLenPar2s = b2(mLenPar2)
    hdr = b"%M$Q" + mID + b1(mDev) + b1(mAct) + mPar0s + mParErrs + mLenPar2s
    mqHdrChk = CalcChkSum(hdr[0:14], False)
    hdr = hdr + b2(mqHdrChk)
    return hdr


def prepareNetHdr(dst, src, blk, type_, blkLen, chkSum):
    blkNum = b2(blk)
    hdr = b1(dst) + b1(src) + b1(blkNum[1]) + b1(blkNum[0])
    hdr = hdr + b1(type_) + b1(blkLen) + b1(chkSum)
    netHdrChk = CalcChkSum(hdr, True)
    hdr = hdr + b1(netHdrChk)
    return hdr


# Was a channel number, instead pass the object
serCh = pyserial.Serial(
    port="/dev/tty.usbmodem123451", baudrate=115200, rtscts=True, timeout=100
)


def prepareNetPayload(fCh, blk, blkLen, fHdr, netSrc, netDst, netType, peerType):
    fHdrLen = len(fHdr)
    if blk == 0:
        netBlk = fHdr + fCh.read(blkLen - fHdrLen)
    else:
        netBlk = fCh.read(blkLen)
    netDataChk = CalcChkSum(netBlk, True)
    netHdr = prepareNetHdr(netDst, netSrc, blk, netType, blkLen, netDataChk)
    netPayload = netHdr + netBlk
    return netPayload


def sendMQMessage(m):
    serCh.write(m)


def waitForACK(mID, timeout):
    returnCode = 0
    while True:
        mQPos = 0
        T_START()
        while True:
            timer = T_COUNT()
            if timer > timeout and mQPos == 0:
                return -1

            nextChar = serCh.read(1)
            if nextChar == b"":
                continue  #  NEXT waitMQSig

            T_START()
            if nextChar[0] != mQSignature[mQPos]:
                mQPos = 0
                continue  # NEXT waitMQSig

            mQHdr[mQPos] = nextChar[0]
            mQPos += 1
            if mQPos >= len(mQSignature):
                break  # EXIT waitMQSig

        for mQHdrPos in range(4, 16):
            while True:
                timer = T_COUNT()
                if timer > timeout and mQHdrPos == 4:
                    returnCode = -2
                    break  # EXIT waitACK

                nextChar = serCh.read(1)
                if nextChar != b"":
                    break  # THEN EXIT waitMQHdr: END IF

            mQHdr[mQHdrPos] = nextChar[0]

        mChkSum = CalcChkSum(mQHdr[0:14], False)
        replyMChkSum = i2(mQHdr[14:16])
        if replyMChkSum != mChkSum:
            returnCode = -3
            break  # EXIT waitACK

        if mID == 116:
            print(replyMChkSum)
            print(mChkSum)
            returnCode = -3
            break

        replyMID = i2(mQHdr[4:6])
        replyErrorFlag = (replyMID & mQErrorMask) != 0
        replyLongFlag = (replyMID & mQLongMask) != 0
        replyReplyFlag = (replyMID & mQReplyMask) != 0

        replyMID = (replyMID & mQMIDMask)
        if replyMID != mID:
            returnCode = -4
            break  # EXIT waitACK

        if replyErrorFlag:
            mParErr = i2(mQHdr[10:12])
            returnCode = -5
            break  # EXIT waitACK

        if not replyReplyFlag:
            returnCode = -6
            break  # EXIT waitACK

        if replyLongFlag:
            mLenPar2 = i2(mQHdr[12:14])
            mQPayload = serCh.read(mLenPar2)
            returnCode = 1
        else:
            returnCode = 0

        break  # EXIT waitACK

    T_STOP()

    return returnCode


def UpdateAndSetNetProfile(peerType):

    mID = getNextMID()
    mDev = 2
    mAct = 4
    mPar0 = 0
    mParErr = peerType
    mLenPar2 = 32
    mQHdr = prepareMQHdr(mID, True, mDev, mAct, mPar0, mParErr, mLenPar2)
    mQPayload = netProfile[peerType]

    sendMQMessage(mQHdr + mQPayload)

    for bb in (mQHdr + mQPayload):
        sys.stdout.write("{0:02x} ".format(bb))

    result = waitForACK(mID, 200)
    if result < 0:
        return result

    print("A returned {0}".format(result))

    mID = getNextMID()
    mDev = 2
    mAct = 5
    mPar0 = 0
    mParErr = peerType
    mLenPar2 = 0
    mQHdr = prepareMQHdr(mID, False, mDev , mAct, mPar0, mParErr, mLenPar2)

    sendMQMessage(mQHdr)
    result = waitForACK(mID, 200)

    print("B returned {0}".format(result))

    return result


def openAndSendFile(fnm, netSrc, netDst, peerType):
    # open file
    fiLen = os.path.getsize(fnm)
    fCh = open(fnm, "rb")

    # peer type
    if peerType == 0:
        # QL peer type = 0
        fiType = 0
        fiDSpace = 0
        fHdr = (
            b"\xff"
            + b4(fiLen)
            + b"\x00"
            + fiType.to_bytes(1, byteorder="big")
        )
        fHdr = fHdr + b4(fiDSpace) + bytearray(4)
        fHdrLen = len(fHdr)
        # fiDSpace = displayFileHdr(fHdr[2:16]) #TODO ?? fHdr(2 TO 15)
    else:
        # ZX Spectrum peer type = 1, read first 9 bytes (?)


# 4560   REMark *** Offset  PROG    #Arr    $Arr    CODE
# 4570   REMark *** 1       0       1       2       3
# 4580   REMark *** 2-3     Len     Len     Len     Len
# 4590   REMark *** 4-5     *       -       -       Start
# 4600   REMark *** 6-7     Var o/s name+32 name+96 -
# 4610   REMark *** 8-9     AutoL** -       -       -

        fHdr = (
            b"\x03" +
            (fiLen).to_bytes(2, byteorder="little") +
            (16384).to_bytes(2, byteorder="little") +
            b"\x00\x00\x00\x00"
        )
        fHdrLen = 9
        # displayFileHdrZX(fHdr)

    # net profile
    result=UpdateAndSetNetProfile(peerType)
    print(result)

    # count blocks
    blkCnt = int((fHdrLen + fiLen) / netBlkLen)
    lastBlkLen = (fHdrLen + fiLen) - (blkCnt * netBlkLen)
    if lastBlkLen > 0:
        lastBlk = blkCnt
    else:
        lastBlk = blkCnt - 1
        lastBlkLen = netBlkLen

    # send all blocks
    for blk in range(lastBlk + 1):
        # net payload
        if blk == lastBlk:
            blkLen = lastBlkLen
            netType = 1
        else:
            blkLen = netBlkLen
            netType = 0

        netPayload = prepareNetPayload(
            fCh, blk, blkLen, fHdr, netSrc, netDst, netType, peerType
        )
        sendPending = 1

        while True:  # REPeat sendBlkAndWaitACK
            # send message
            if sendPending:
                mID = getNextMID()
                print("send    {0} {1}".format(blk, mID))
                mDev = 2
                mAct = 3
                mPar0 = 0
                mParErr = 0
                mLenPar2 = len(netPayload)
                nextMQHdr = prepareMQHdr(mID, True, mDev, mAct, mPar0, mParErr, mLenPar2)
                sendMQMessage(nextMQHdr + netPayload)
                sendPending = 0

            serCh.flush()

            # ack message
            result = waitForACK(mID, 500)

            # handle result
            if result == 0:
                print("ack     {0} {1}".format(blk, mID))
                # *** OK - Send next packet
                break  # EXIT sendBlkAndWaitACK
            elif result == -1:
                print("timeout {0} {1}".format(blk, mID))
                # *** Timeout - not an error as such, keep waiting...
                continue  # NEXT sendBlkAndWaitACK
            elif result == -2:
                print("timeout {0} {1} after receiving MQSYNC".format(blk, mID))
                res = "Timeout after receiving MQSYNC"
            elif result == -3:
                print("chksum  {0} {1}".format(blk, mID))
                res = "Mismatched Chksum"
                sendPending = 1
            elif result == -4:
                print("mid     {0} {1}".format(blk, mID))
                res = "MID wasn't as expected"
            elif result == -5:
                print("err     {0} {1}".format(blk, mID))
                res = "ERRor Flag set in response"
                sendPending = 1
            elif result == -6:
                print("rep     {0} {1}".format(blk, mID))
                res = "REPly Flag not set in response"
            else:
                print("result  {0} {1}".format(blk, mID))
                res = "Unknown"

            print("Unexpected MQ error:")
            print(result)
            print(res)

    fCh.close()

if len(sys.argv) != 5:
    print("Usage: %s <profile> <source> <dest> <file>" % sys.argv[0])
    sys.exit(1)

openAndSendFile(sys.argv[4], int(sys.argv[2]), int(sys.argv[3]), int(sys.argv[1]))
